# main 74
# sam robson

import click
import pandas as pd
import sys
import warnings
from alacorder import logs
from alacorder import get
from alacorder import parse
from alacorder import write
from alacorder import config

pd.set_option('display.max_rows', 1000)

@click.command()
@click.option('--input-path','-in',required=True,type=click.Path(), prompt=logs.title(),help="Path to input archive or PDF directory", show_choices=False)
@click.option('--output-path','-out',prompt=logs.both(),type=click.Path(), help="Path to output table (.xls, .xlsx, .csv, .json, .dta) or archive (.pkl.xz)", show_choices=False)
@click.option('--count','-c',default=0, help='Max cases to pull from input',show_default=False)
@click.option('--table','-t', help="Table export choice")
@click.option('--overwrite', '-o', default=False, help="Overwrite output path if exists", is_flag=True, show_default=False)
@click.option('--launch', default=False, is_flag=True, help="Launch export in default application", show_default=False)
@click.option('--dedupe','-dd', default=False, is_flag=True, help="Remove duplicate cases from input archive",hidden=True)
@click.option('--log / --no-log', default=True, is_flag=True, help="Print outputs to console upon completion")
@click.option('--no-write', default=False, is_flag=True, help="Do not export to output path",hidden=True)
@click.option('--no-prompt','-np', default=False, is_flag=True, help="Skip confirmation prompts")
@click.option('--debug', default=False, is_flag=True, help="Prints extensive logs to console for development purposes")
@click.option('--no-batch', default=False,is_flag=True,help="Process all inputs as one batch")
@click.option('--warn', default=False,is_flag=True,help="Process all inputs as one batch",hidden=True)
def cli(input_path, output_path, count, table, overwrite, launch, dedupe, log, no_write, no_prompt, debug, no_batch, warn):

	show_options_menu = True if table == None and no_prompt == False and count == 0 and overwrite == False and launch == False and dedupe == False and log == True and no_write == False and no_prompt == False and debug == False and no_batch == False else False

	# suppress tracebacks unless debug
	if not debug:
		sys.tracebacklimit = 0
	if not warn:
		warnings.filterwarnings('ignore')

	# inputs - configure and log
	inputs = config.inputs(input_path)
	if debug:
		click.echo(inputs)
	if log:
		click.echo(inputs.ECHO)
	if not inputs.GOOD:
		raise Exception("Invalid input path!")

	# outputs - configure and log
	outputs = config.outputs(output_path)
	if debug:
		click.echo(outputs)
	if log:
		click.echo(outputs.ECHO)
	if not outputs.GOOD:
		raise Exception("Invalid input path!")

	# prompt overwrite 
	if outputs.EXISTING_FILE and not overwrite:
		if no_prompt:
			raise Exception("Existing file at output path! Repeat with flag --overwrite to replace file.")
		else:
			if click.confirm(logs.echo_yellow("Existing file at output path will be written over! Continue?",echo=False)):
				pass
			else:
				raise Exception("Existing file at output path!")

	# prompt table 
	if outputs.MAKE == "multiexport" and table != "cases" and table != "fees" and table != "charges" and table != "disposition" and table != "filing":
		table = "all"
	if outputs.MAKE == "singletable" and table != "cases" and table != "fees" and table != "charges" and table != "disposition" and table != "filing":
		if no_prompt:
			raise Exception("Invalid/missing table selection!")
		else: 
			pick = click.prompt(logs.pick_table()) # add str
			if pick == "A":
				table = "cases"
			elif pick == "B":
				table = "fees"
			elif pick == "C":
				table = "charges"
			elif pick == "D":
				table = "disposition"
			elif pick == "E":
				table = "filing"
			else:
				raise Exception("Invalid table selection!")

	# prompt options
	if show_options_menu and not no_prompt:
		if not click.confirm("Continue with current settings? "):
			cli.main(['alacorder','--help'],standalone_mode=False)
			p = click.prompt('Enter the <option> flag you would like to set: ')
			if p == "count" or p == "-c" or p == "--count":
				count = click.prompt("Set max case count to pull from input: ",type=int)
			elif p == "overwrite" or p == "--overwrite" or p == "-o":
				overwrite = click.prompt("Should Alacorder OVERWRITE existing files at provided output file paths? [y/N]",type=bool)
			elif p == "launch" or p == "--launch":
				launch = click.prompt("Should Alacorder attempt to launch exported files once complete? [y/N]",type=bool)
			elif p == "dedupe" or p == "--dedupe" or p == "-dd":
				dedupe = click.prompt("Should Alacorder attempt to remove duplicate cases from outputs? [y/N]",type=bool) # might change to just table
			elif p == "log" or p == "--log" or p == "no-log" or p == "--no-log":
				log = click.prompt("Should Alacorder print logs to console? [y/N]",type=bool) # might change to just table
			elif p == "no_prompt" or p == "--no-prompt" or p == "-np":
				no_prompt = click.prompt("Should Alacorder proceed without prompting for user input? [y/N]",type=bool) # might change to just table
			elif p == "debug" or p == "--debug" or p == "-d":
				debug = click.prompt("Should Alacorder print detailed debug logs? [y/N]",type=bool) # might change to just table
			elif p == "no_batch" or p == "--no-batch":
				no_batch = click.prompt("Should Alacorder process all cases in one batch? [y/N]",type=bool) # might change to just table
			else:
				click.echo("Option not found.")
			if debug:
				click.echo(p)

	# finalize config
	cf = config.set(inputs, outputs, count=count, table=table, overwrite=overwrite, launch=launch, log=log, dedupe=dedupe, no_write=no_write, no_prompt=no_prompt, no_batch=no_batch, debug=debug)

	if debug:
		click.echo(cf)
	if log:
		click.echo(cf.ECHO)

	if cf.MAKE == "archive":
		o = write.archive(cf)
		logs.debug(cf,o.describe())
	if cf.MAKE == "multiexport" and cf.TABLE == "all":
		o = parse.cases(cf)
		logs.debug(cf,o[0].describe(),o[1].describe(),o[2].describe())
	if cf.TABLE == "fees":
		o = parse.fees(cf)
		logs.debug(cf,o.describe())
	if cf.TABLE == "charges" or cf.TABLE == "disposition" or cf.TABLE == "filing":
		o = parse.charges(cf) 
		logs.debug(cf,o.describe())
	if cf.TABLE == "cases":
		o = parse.caseinfo(cf)
		logs.debug(cf,o.describe())

	
	logs.echo(cf,"Alacorder completed and quit.")


if __name__ == "__main__":
	cli()





# log 74
# Sam Robson

import os
import sys
import glob
import re
import math
import xarray
import numpy as np
import datetime
import pandas as pd
import time
import warnings
import click
import inspect
from alacorder import get
from alacorder import parse
from alacorder import write
from alacorder import config

def echo_conf(input_path,make,output_path,overwrite,no_write,dedupe,launch,warn,no_prompt):
	d = click.style(f"""\n* Successfully configured!\n""",fg='green', bold=True)
	e = click.style(f"""INPUT: {input_path}\n{'TABLE' if make == "multiexport" or make == "singletable" else 'ARCHIVE'}: {output_path}\n""",fg='bright_yellow',bold=True)
	f = click.style(f"""{"OVERWRITE is enabled. Alacorder will overwrite existing files at output path! " if overwrite else ''}{"NO-WRITE is enabled. Alacorder will NOT export outputs. " if no_write else ''}{"REMOVE DUPLICATES is enabled. At time of export, all duplicate cases will be removed from output. " if dedupe else ''}{"LAUNCH is enabled. Upon completion, Alacorder will attempt to launch exported file in default viewing application. " if launch and make != "archive" else ''}{"WARN is enabled. All warnings from pandas and other modules will print to console. " if warn else ''}{"NO_PROMPT is enabled. All user confirmation prompts will be suppressed as if set to default by user." if no_prompt else ''}""".strip(), italic=True, fg='white')
	return d + e + "\n" + f + "\n"

def complete(conf, *outputs):
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	if conf.LOG and len(outputs)>0:
		click.echo(outputs)
	if conf.LOG:
		click.secho(f'''\n* Task completed!\n''',bold=True,fg='green')

def debug(conf, *msg):
	if conf['DEBUG'] == True:
		click.echo(msg)

def echo(conf, *msg):
	if conf['LOG']==True:
		click.echo(msg)

def echo_red(text, echo=True):
	if echo:
		click.echo(click.style(text,fg='bright_red',bold=True),nl=True)
		return click.style(text,fg='bright_red',bold=True)
	else:
		return click.style(text,fg='bright_red',bold=True)
def echo_yellow(text, echo=True):
	if echo:
		click.echo(click.style(text,fg='bright_yellow',bold=True),nl=True)
		return click.style(text,fg='bright_yellow',bold=True)
	else:
		return click.style(text,fg='bright_yellow',bold=True)
def echo_green(text, echo=True):
	if echo:
		click.echo(click.style(text,fg='bright_green',bold=True),nl=True)
		return click.style(text,fg='bright_green',bold=True)
	else:
		return click.style(text,fg='bright_green',bold=True)

upick_table = ('''
Select preferred table output below.
	A:  Case Details
	B:  Fee Sheets
	C:  Charges (all)
	D:  Charges (disposition only)
	E:  Charges (filing only)

>> Enter A, B, C, D, or E to continue:

''')
def pick_table():
	return click.style(upick_table,bold=True)
ujust_table = ('''

EXPORT DATA TABLE: To export data table from case inputs, enter full output path. Use .xls or .xlsx to export all tables, or, if using another format (.csv, .json, .dta), select a table after entering output file path.

>> Enter path:

''')

def just_table():
	return click.style(ujust_table,bold=True)
uboth =  ('''

EXPORT FULL TEXT ARCHIVE: To process case inputs into a full text archive (recommended), enter archive path below with file extension .pkl.xz.

EXPORT DATA TABLE: To export data table from case inputs, enter full output path. Use .xls or .xlsx to export all tables, or, if using another format (.csv, .json, .dta), select a table after entering output file path.

>> Enter path:

''')

def both():
	return click.style(uboth,fg='bright_white')
utitle = ('''

ALACORDER beta 74
© 2023 Sam Robson

Alacorder processes case detail PDFs into data tables suitable for research purposes. Alacorder also generates compressed text archives from the source PDFs to speed future data collection from the same set of cases.

ACCEPTED      /pdfs/path/   PDF directory           
INPUTS:       .pkl.xz       Compressed archive      

>> Enter input path: 

''')

def title():
	return click.style(utitle,fg='bright_white')

utext_p = ('''

>> Enter path to output text file (must be .txt): 

''')

def text_p():
	return click.style(utext_p,bold=True)








# parse 74
# Sam Robson

import os
import sys
import glob
import re
import math
import numexpr
import xarray
import bottleneck
import numpy as np
import xlrd
import openpyxl
import datetime
import pandas as pd
import time
import warnings
import click
import inspect
from alacorder import get 
from alacorder import write 
from alacorder import config 
from alacorder import logs
import PyPDF2
from io import StringIO
try:
	import xlsxwriter
except ImportError:
	pass
def table(conf):
	"""
	Route config to parse...() function corresponding to table attr 
	"""
	a = []
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	if conf.MAKE == "multiexport":
		a = cases(conf)
	if conf['TABLE'] == "cases":
		a = caseinfo(conf)
	if conf['TABLE'] == "fees":
		a = fees(conf)
	if conf['TABLE'] == "charges":
		a = charges(conf)
	if conf['TABLE'] == "disposition":
		a = charges(conf)
	if conf['TABLE'] == "filing":
		a = charges(conf)
	return a

def fees(conf):
	"""
	Return fee sheets with case number as DataFrame from batch
	fees = pd.DataFrame({'CaseNumber': '', 
		'Code': '', 'Payor': '', 'AmtDue': '', 
		'AmtPaid': '', 'Balance': '', 'AmtHold': ''})
	"""
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	from_archive = True if conf['IS_FULL_TEXT'] else False
	start_time = time.time()
	if warn == False:
		warnings.filterwarnings("ignore")
	fees = pd.DataFrame()
	# fees = pd.DataFrame({'CaseNumber': '', 
	  #  'Code': '', 'Payor': '', 'AmtDue': '', 
	   # 'AmtPaid': '', 'Balance': '', 'AmtHold': ''},index=[0])
	if not conf['NO_BATCH']:
		batches = config.batcher(conf)
	else:
		batches = np.array_split(queue, 1)
	batchsize = max(pd.Series(batches).map(lambda x: x.shape[0]))
	batchcount = len(batches)
	with click.progressbar(batches) as bar:
		for i, c in enumerate(bar):
			exptime = time.time()
			b = pd.DataFrame()

			if from_archive == True:
				b['AllPagesText'] = c
			else:
				b['AllPagesText'] = c.map(lambda x: get.PDFText(x))

			b['CaseInfoOutputs'] = b['AllPagesText'].map(lambda x: get.CaseInfo(x))
			b['CaseNumber'] = b['CaseInfoOutputs'].map(lambda x: x[0])
			#try:
			b['FeeOutputs'] = b.index.map(lambda x: get.FeeSheet(str(b.loc[x].AllPagesText)))
			feesheet = b['FeeOutputs'].map(lambda x: x[6]) 
			#except (AttributeError,IndexError):
			 #   pass

			feesheet = feesheet.dropna() # drop empty 
			fees =fees.dropna()
			feesheet = feesheet.tolist() # convert to list -> [df, df, df]
			#try:
			feesheet = pd.concat(feesheet,axis=0,ignore_index=True) # add all dfs in batch -> df
			#except ValueError:
			 #   pass
			fees = fees.append(feesheet, ignore_index=True) 
			fees.fillna('',inplace=True)
			fees['AmtDue'] = fees['AmtDue'].map(lambda x: pd.to_numeric(x,'coerce'))
			fees['AmtPaid'] = fees['AmtPaid'].map(lambda x: pd.to_numeric(x,'coerce'))
			fees['Balance'] = fees['Balance'].map(lambda x: pd.to_numeric(x,'coerce'))
			fees['AmtHold'] = fees['AmtHold'].map(lambda x: pd.to_numeric(x,'coerce'))
	if not no_write:
		write.now(conf, fees)
	logs.complete(conf)
	return fees
def charges(conf):
	"""
	Return charges with case number as DataFrame from batch
	charges = pd.DataFrame({'CaseNumber': '', 'Num': '', 'Code': '', 'Felony': '', 'Conviction': '', 'CERV': '', 'Pardon': '', 'Permanent': '', 'Disposition': '', 'CourtActionDate': '', 'CourtAction': '', 'Cite': '', 'TypeDescription': '', 'Category': '', 'Description': ''}) 
	"""
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	dedupe = conf['DEDUPE']
	from_archive = True if conf['IS_FULL_TEXT'] else False

	if warn == False:
		warnings.filterwarnings("ignore")

	if not conf['NO_BATCH']:
		batches = config.batcher(conf)
	else:
		batches = np.array_split(queue, 1)
	batchsize = max(pd.Series(batches).map(lambda x: x.shape[0]))

	start_time = time.time()
	charges = pd.DataFrame()
	with click.progressbar(batches) as bar:
		for i, c in enumerate(bar):
			exptime = time.time()
			b = pd.DataFrame()

			if from_archive == True:
				b['AllPagesText'] = c
			else:
				b['AllPagesText'] = pd.Series(c).map(lambda x: get.PDFText(x))

			## b['CaseInfoOutputs'] = b['AllPagesText'].map(lambda x: get.CaseInfo(x))
			b['CaseNumber'] = b['AllPagesText'].map(lambda x: get.CaseNumber(x))
			b['ChargesOutputs'] = b.index.map(lambda x: get.Charges(str(b.loc[x].AllPagesText)))

			
			chargetabs = b['ChargesOutputs'].map(lambda x: x[17])
			chargetabs = chargetabs.dropna()
			chargetabs = chargetabs.tolist()
			charges = charges.append(chargetabs)
			charges.fillna('',inplace=True)

		if table == "filing":
			is_disp = charges['Disposition']
			is_filing = is_disp.map(lambda x: False if x == True else True)
			charges = charges[is_filing]
			charges.drop(columns=['CourtAction','CourtActionDate'],inplace=True)

		if table == "disposition":
			is_disp = charges.Disposition.map(lambda x: True if x == True else False)
			charges = charges[is_disp]
		if not no_write:
			write.now(conf, charges)

	logs.complete(conf)
	return charges
def cases(conf):
	"""
	Return [cases, fees, charges] tables as List of DataFrames from batch
	See API docs for table specific outputs
	"""
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	overwrite = conf['OVERWRITE']
	path_out = conf['OUTPUT_PATH'] if conf.MAKE != "archive" else ''
	archive_out = conf['OUTPUT_PATH'] if conf.MAKE == "archive" else ''
	from_archive = True if conf['IS_FULL_TEXT'] else False
	start_time = time.time()
	arc_ext = conf['OUTPUT_EXT']
	cases = pd.DataFrame()
	fees = pd.DataFrame()
	charges = pd.DataFrame()
	if not conf['NO_BATCH']:
		batches = config.batcher(conf)
	else:
		batches = np.array_split(queue, 1)
	batchsize = max(pd.Series(batches).map(lambda x: x.shape[0]))
	if warn == False:
		warnings.filterwarnings("ignore")
	temp_no_write_arc = False
	temp_no_write_tab = False
	with click.progressbar(batches) as bar:
		for i, c in enumerate(bar):
			b = pd.DataFrame()
			if from_archive == True:
				b['AllPagesText'] = c
			else:
				b['AllPagesText'] = pd.Series(c).map(lambda x: get.PDFText(x))
			b['CaseInfoOutputs'] = b['AllPagesText'].map(lambda x: get.CaseInfo(x))
			b['CaseNumber'] = b['CaseInfoOutputs'].map(lambda x: x[0])
			b['Name'] = b['CaseInfoOutputs'].map(lambda x: x[1])
			b['Alias'] = b['CaseInfoOutputs'].map(lambda x: x[2])
			b['DOB'] = b['CaseInfoOutputs'].map(lambda x: x[3])
			b['Race'] = b['CaseInfoOutputs'].map(lambda x: x[4])
			b['Sex'] = b['CaseInfoOutputs'].map(lambda x: x[5])
			b['Address'] = b['CaseInfoOutputs'].map(lambda x: x[6])
			b['Phone'] = b['CaseInfoOutputs'].map(lambda x: x[7])
			b['ChargesOutputs'] = b.index.map(lambda x: get.Charges(str(b.loc[x].AllPagesText)))
			b['Convictions'] = b['ChargesOutputs'].map(lambda x: x[0])
			b['Dispositioncharges'] = b['ChargesOutputs'].map(lambda x: x[1])
			b['FilingCharges'] = b['ChargesOutputs'].map(lambda x: x[2])
			b['CERVConvictions'] = b['ChargesOutputs'].map(lambda x: x[3])
			b['PardonConvictions'] = b['ChargesOutputs'].map(lambda x: x[4])
			b['PermanentConvictions'] = b['ChargesOutputs'].map(lambda x: x[5])
			b['ConvictionCount'] = b['ChargesOutputs'].map(lambda x: x[6])
			b['ChargeCount'] = b['ChargesOutputs'].map(lambda x: x[7])
			b['CERVChargeCount'] = b['ChargesOutputs'].map(lambda x: x[8])
			b['PardonChargeCount'] = b['ChargesOutputs'].map(lambda x: x[9])
			b['PermanentChargeCount'] = b['ChargesOutputs'].map(lambda x: x[10])
			b['CERVConvictionCount'] = b['ChargesOutputs'].map(lambda x: x[11])
			b['PardonConvictionCount'] = b['ChargesOutputs'].map(lambda x: x[12])
			b['PermanentConvictionCount'] = b['ChargesOutputs'].map(lambda x: x[13])
			b['ChargeCodes'] = b['ChargesOutputs'].map(lambda x: x[14])
			b['ConvictionCodes'] = b['ChargesOutputs'].map(lambda x: x[15])
			b['FeeOutputs'] = b.index.map(lambda x: get.FeeSheet(str(b.loc[x].AllPagesText)))
			b['TotalAmtDue'] = b['FeeOutputs'].map(lambda x: x[0])
			b['TotalBalance'] = b['FeeOutputs'].map(lambda x: x[1])
			b['PaymentToRestore'] = b['AllPagesText'].map(lambda x: get.PaymentToRestore(x))
			# b['PaymentToRestore'][b['CERVConvictionCount'] == 0] = pd.NaT
			b['FeeCodesOwed'] = b['FeeOutputs'].map(lambda x: x[3])
			b['FeeCodes'] = b['FeeOutputs'].map(lambda x: x[4])
			b['FeeSheet'] = b['FeeOutputs'].map(lambda x: x[5])
			logs.debug(conf, b['FeeSheet'])

			feesheet = b['FeeOutputs'].map(lambda x: x[6]) 
			feesheet = feesheet.dropna() 
			feesheet = feesheet.tolist() # -> [df, df, df]
			
			#try:
			feesheet = pd.concat(feesheet,axis=0,ignore_index=True) #  -> batch df
			#except:
			 #   pass
			#try:
			fees = fees.append(feesheet)
			#except:
			 #   pass
			logs.debug(conf, fees)
			chargetabs = b['ChargesOutputs'].map(lambda x: x[17])
			chargetabs = chargetabs.dropna()
			# charges = charges.dropna()
			chargetabs = chargetabs.tolist()
			
			#try:
			chargetabs = pd.concat(chargetabs,axis=0,ignore_index=True)
			#except:
			 #   pass
			#try:
			charges = charges.append(chargetabs,ignore_index=True)
			#except:
			 #   pass
			
			feesheet['AmtDue'] = feesheet['AmtDue'].map(lambda x: pd.to_numeric(x,'coerce'))
			feesheet['AmtPaid'] = feesheet['AmtPaid'].map(lambda x: pd.to_numeric(x,'coerce'))
			feesheet['Balance'] = feesheet['Balance'].map(lambda x: pd.to_numeric(x,'coerce'))
			feesheet['AmtHold'] = feesheet['AmtHold'].map(lambda x: pd.to_numeric(x,'coerce'))

			b['ChargesTable'] = b['ChargesOutputs'].map(lambda x: x[-1])
			b['Phone'] =  b['Phone'].map(lambda x: pd.to_numeric(x,'coerce'))
			b['TotalAmtDue'] = b['TotalAmtDue'].map(lambda x: pd.to_numeric(x,'coerce'))
			b['TotalBalance'] = b['TotalBalance'].map(lambda x: pd.to_numeric(x,'coerce'))
			b['PaymentToRestore'] = b['TotalBalance'].map(lambda x: pd.to_numeric(x,'coerce'))

			if bool(archive_out) and len(arc_ext) > 2 and i > 0 and not no_write:
				if os.path.getsize(archive_out) > 1000:
					temp_no_write_arc = True
			if bool(path_out) and i > 0 and not no_write:
				if os.path.getsize(path_out) > 1000:
					temp_no_write_tab = True
			if i == len(batches) - 1:
				temp_no_write_arc = False
				temp_no_write_tab = False

			if (i % 5 == 0 or i == len(batches) - 1) and not no_write and temp_no_write_arc == False:
				if bool(archive_out) and len(arc_ext) > 2:
					timestamp = start_time
					q = pd.Series(queue) if conf.IS_FULL_TEXT == False else pd.NaT
					ar = pd.DataFrame({
						'Path': q,
						'AllPagesText': b['AllPagesText'],
						'Timestamp': timestamp
						},index=range(0,conf.COUNT))
					try:
						arch = pd.concat([arch, ar],ignore_index=True,axis=0)
					except:
						pass
					arch.fillna('',inplace=True)
					arch.dropna(inplace=True)
					arch.to_pickle(archive_out,compression="xz")

			b.drop(columns=['AllPagesText','CaseInfoOutputs','ChargesOutputs','FeeOutputs','ChargesTable','FeeSheet'],inplace=True)

			if dedupe == True:
				outputs.drop_duplicates(keep='first',inplace=True)
			
			b.fillna('',inplace=True)
			# newcases = [cases, b]
			cases = cases.append(b, ignore_index=True)

			if no_write == False and temp_no_write_tab == False and (i % 5 == 0 or i == len(batches) - 1):
				if out_ext == ".xls":
					try:
						with pd.ExcelWriter(path_out,engine="openpyxl") as writer:
							cases.to_excel(writer, sheet_name="cases")
							fees.to_excel(writer, sheet_name="fees")
							charges.to_excel(writer, sheet_name="charges")
					except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
						click.echo(f"openpyxl engine failed! Trying xlsxwriter...")
						with pd.ExcelWriter(path_out,engine="xlsxwriter") as writer:
							cases.to_excel(writer, sheet_name="cases")
							fees.to_excel(writer, sheet_name="fees")
							charges.to_excel(writer, sheet_name="charges")
				elif out_ext == ".xlsx":
					try:
						with pd.ExcelWriter(path_out,engine="openpyxl") as writer:
							cases.to_excel(writer, sheet_name="cases")
							fees.to_excel(writer, sheet_name="fees")
							charges.to_excel(writer, sheet_name="charges")
					except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
						try:
							if warn:
								click.echo(f"openpyxl engine failed! Trying xlsxwriter...")
							with pd.ExcelWriter(path_out,engine="xlsxwriter") as writer:
								cases.to_excel(writer, sheet_name="cases")
								fees.to_excel(writer, sheet_name="fees")
								charges.to_excel(writer, sheet_name="charges")
						except (ImportError, FileNotFoundError, IndexError, ValueError, ModuleNotFoundError):
							try:
								cases.to_json(os.path.splitext(path_out)[0] + "-cases.json.zip", orient='table')
								fees.to_json(os.path.splitext(path_out)[0] + "-fees.json.zip",orient='table')
								charges.to_json(os.path.splitext(path_out)[0] + "-charges.json.zip",orient='table')
								logs.echo(conf,"Fallback export to " + os.path.splitext(path_out)[0] + "-cases.json.zip due to Excel engine failure, usually caused by exceeding max row limit for .xls/.xlsx files!")
							except (ImportError, FileNotFoundError, IndexError, ValueError):
								click.echo("Failed to export!")
							
				elif out_ext == ".json":
					cases.to_json(path_out,orient='table')
				elif out_ext == ".csv":
					cases.to_csv(path_out,escapechar='\\')
				elif out_ext == ".md":
					cases.to_markdown(path_out)
				elif out_ext == ".txt":
					cases.to_string(path_out)
				elif out_ext == ".dta":
					cases.to_stata(path_out)
				elif out_ext == ".parquet":
					outputs.to_parquet(path_out)
				else:
					pd.Series([cases, fees, charges]).to_string(path_out)
				try:
					if dedupe == True and outputs.shape[0] < queue.shape[0]:
						click.echo(f"Identified and removed {outputs.shape[0]-queue.shape[0]} from queue.")
				except:
					pass

		logs.complete(conf)
		return [cases, fees, charges]

def caseinfo(conf):
	"""
	Return case information with case number as DataFrame from batch
	List: ['CaseNumber','Name','Alias','DOB','Race','Sex','Address','Phone','Totals','TotalAmtDue','TotalAmtPaid','TotalBalance','TotalAmtHold','PaymentToRestore','ConvictionCodes','ChargeCodes','FeeCodes','FeeCodesOwed','Dispositioncharges','FilingCharges','CERVConvictions','PardonDQConvictions','PermanentDQConviction','TotalAmtDue','TotalAmtPaid','TotalBalance','TotalAmtHold','PaymentToRestore','ConvictionCodes','ChargeCodes','FeeCodes','FeeCodesOwed','Dispositioncharges','FilingCharges','CERVConvictions','PardonDQConvictions','PermanentDQConvictions']
	"""
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	dedupe = conf['DEDUPE']
	path_out = conf['OUTPUT_PATH'] if conf.MAKE != "archive" else ''
	archive_out = conf['OUTPUT_PATH'] if conf.MAKE == "archive" else ''
	from_archive = True if conf['IS_FULL_TEXT'] else False

	cases = pd.DataFrame()

	if not conf['NO_BATCH']:
		batches = config.batcher(conf)
	else:
		batches = np.array_split(queue, 1)
	batchsize = max(pd.Series(batches).map(lambda x: x.shape[0]))
	logs.debug(conf, [batches,batchsize])
	

	if warn == False:
		warnings.filterwarnings("ignore")
	with click.progressbar(batches) as bar:
		for i, c in enumerate(bar):
			b = pd.DataFrame()
			if from_archive == True:
				b['AllPagesText'] = c
			else:
				b['AllPagesText'] = pd.Series(c).map(lambda x: get.PDFText(x))

			b['CaseInfoOutputs'] = b['AllPagesText'].map(lambda x: get.CaseInfo(x))
			logs.debug(conf, b['CaseInfoOutputs'])
			b['CaseNumber'] = b['CaseInfoOutputs'].map(lambda x: x[0])
			b['Name'] = b['CaseInfoOutputs'].map(lambda x: x[1])
			b['Alias'] = b['CaseInfoOutputs'].map(lambda x: x[2])
			b['DOB'] = b['CaseInfoOutputs'].map(lambda x: x[3])
			b['Race'] = b['CaseInfoOutputs'].map(lambda x: x[4])
			b['Sex'] = b['CaseInfoOutputs'].map(lambda x: x[5])
			b['Address'] = b['CaseInfoOutputs'].map(lambda x: x[6])
			b['Phone'] = b['CaseInfoOutputs'].map(lambda x: x[7])
			b['Totals'] = b['AllPagesText'].map(lambda x: get.Totals(x))
			b['TotalAmtDue'] = b['Totals'].map(lambda x: x[1])
			b['TotalAmtPaid'] = b['Totals'].map(lambda x: x[2])
			b['TotalBalance'] = b['Totals'].map(lambda x: x[3])
			b['TotalAmtHold'] = b['Totals'].map(lambda x: x[4])
			b['PaymentToRestore'] = b['AllPagesText'].map(lambda x: get.PaymentToRestore(x))
			# b['PaymentToRestore'][b['CERVConvictionCount'] == 0] = pd.NaT
			b['ConvictionCodes'] = b['AllPagesText'].map(lambda x: get.ConvictionCodes(x))
			b['ChargeCodes'] = b['AllPagesText'].map(lambda x: get.ChargeCodes(x))
			b['FeeCodes'] = b['AllPagesText'].map(lambda x: get.FeeCodes(x))
			b['FeeCodesOwed'] = b['AllPagesText'].map(lambda x: get.FeeCodesOwed(x))
			b['Dispositioncharges'] = b['AllPagesText'].map(lambda x: get.DispositionCharges(x))
			b['FilingCharges'] = b['AllPagesText'].map(lambda x: get.FilingCharges(x))
			b['CERVConvictions'] = b['AllPagesText'].map(lambda x: get.CERVConvictions(x))
			b['PardonDQConvictions'] = b['AllPagesText'].map(lambda x: get.PardonDQConvictions(x))
			b['PermanentDQConvictions'] = b['AllPagesText'].map(lambda x: get.PermanentDQConvictions(x))
			b['Phone'] =  b['Phone'].map(lambda x: pd.to_numeric(x,'coerce'))
			b['TotalAmtDue'] = b['TotalAmtDue'].map(lambda x: pd.to_numeric(x,'coerce'))
			b['TotalBalance'] = b['TotalBalance'].map(lambda x: pd.to_numeric(x,'coerce'))
			b.drop(columns=['AllPagesText','CaseInfoOutputs','Totals'],inplace=True)
			b.fillna('',inplace=True)
			# newcases = [cases, b]
			cases = cases.append(b, ignore_index=True)
			# write 
		if not no_write:
			write.now(conf, cases)
		logs.complete(conf)
		return cases
def map(conf, *args):
	"""
	Custom Parsing
	From config object and custom getter functions defined like below:

	def getter(text: str):
		out = re.search(...)
		...
		return str(out)

	Creates DataFrame with column for each getter column output and row for each case in queue

	"""
	if not conf.DEBUG:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	logs.debug(conf,conf)
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	path_out = conf['OUTPUT_PATH'] if conf.MAKE != "archive" else ''
	archive_out = conf['OUTPUT_PATH'] if conf.MAKE == "archive" else ''
	from_archive = True if conf['IS_FULL_TEXT']==True else False

	if warn == False:
		warnings.filterwarnings("ignore")
	if not conf['NO_BATCH']:
		batches = config.batcher(conf)
	else:
		batches = np.array_split(queue, 1)
	batchsize = max(pd.Series(batches).map(lambda x: x.shape[0]))
	logs.debug(conf, [batches, batchsize])
	start_time = time.time()
	alloutputs = []
	uselist = False
	func = pd.Series(args).map(lambda x: 1 if inspect.isfunction(x) else 0)
	funcs = func.index.map(lambda x: args[x] if func[x]>0 else np.nan)
	no_funcs = func.index.map(lambda x: args[x] if func[x]==0 else np.nan)
	no_funcs = no_funcs.dropna()
	countfunc = func.sum()
	column_getters = pd.DataFrame(columns=['Name','Method','Arguments'],index=(range(0,countfunc)))
	df_out = pd.DataFrame()
	local_get = []
	for i, x in enumerate(funcs):
		if inspect.isfunction(x):
			column_getters.Name[i] = x.__name__
			column_getters.Method[i] = x
	for i, x in enumerate(args):
		if inspect.isfunction(x) == False:
			column_getters.Arguments.iloc[i-1] = x
	if print_log:
		click.echo(column_getters)
	def ExceptionWrapperArgs(mfunc, x, *args):
		unpacked_args = args
		a = mfunc(x, unpacked_args)
		return a
	def ExceptionWrapper(mfunc, x):
		a = str(mfunc(x))
		return a
	temp_no_write_tab = False
	with click.progressbar(batches) as bar:
		for i, c in enumerate(bar):
			exptime = time.time()
			b = pd.DataFrame()

			if bool(path_out) and i > 0 and not no_write:
				if os.path.getsize(path_out) > 500:
					temp_no_write_tab = True
			if i == len(batches) - 1:
				temp_no_write_tab = False
			if from_archive == True:
				allpagestext = c
			else:
				allpagestext = pd.Series(c).map(lambda x: get.PDFText(x))
			df_out['CaseNumber'] = allpagestext.map(lambda x: get.CaseNumber(x))
			for i, getter in enumerate(column_getters.Method.tolist()):
				arg = column_getters.Arguments[i]
				try:
					name = getter.__name__.strip()
					col = pd.DataFrame({
					name: allpagestext.map(lambda x: getter(x, arg))
						})
				except (AttributeError,TypeError):
					try:
						name = getter.__name__.strip()
						col = pd.DataFrame({
						name: allpagestext.map(lambda x: getter(x))
								})
					except (AttributeError,TypeError):
						name = getter.__name__.strip()
						col = pd.DataFrame({
						name: allpagestext.map(lambda x: ExceptionWrapper(x,arg))
								})
				n_out = [df_out, col]
				df_out = pd.concat([df_out,col.reindex(df_out.index)],axis=1)
				df_out = df_out.dropna(axis=1)
				df_out = df_out.convert_dtypes()
				logs.debug(conf, df_out)
			if no_write == False and temp_no_write_tab == False and (i % 5 == 0 or i == len(batches) - 1):
				write.now(conf, df_out) # rem alac
	if not no_write:
		write.now(conf, df_out) # rem alac
	logs.complete(conf, start_time)
	return df_out



# write 74
# Sam Robson

import os
import sys
import glob
import re
import math
import xarray
import numpy as np
import datetime
import pandas as pd
import time
import warnings
import click
import inspect
from alacorder import get 
from alacorder import parse 
from alacorder import config 
from alacorder import logs

def now(conf, outputs, archive=False):
	"""
	Writes outputs to path in conf
	"""
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	arc_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	dedupe = conf['DEDUPE']
	launch = conf['LAUNCH']
	path_out = conf['OUTPUT_PATH'] if conf['MAKE'] != "archive" else ''
	archive_out = conf['OUTPUT_PATH'] if conf['MAKE'] == "archive" else ''
	from_archive = True if conf['IS_FULL_TEXT']==True else False

	try:
		out_ext = os.path.splitext(path_out)[1]
	except TypeError:
		out_ext = ""

	if out_ext == ".xls":
		try:
			with pd.ExcelWriter(path_out) as writer:
				outputs.to_excel(writer, sheet_name="output-table",engine="openpyxl")
		except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
			try:
				with pd.ExcelWriter(path_out,engine="xlwt") as writer:
					outputs.to_excel(writer, sheet_name="output-table")
			except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
				try:
					os.remove(path_out)
				except:
					pass
				outputs.to_json(os.path.splitext(path_out)[0] + "-cases.json.zip", orient='table')
				if warn or print_log:
					click.echo(f"Fallback export to {os.path.splitext(path_out)[0]}-cases.json.zip due to Excel engine failure, usually caused by exceeding max row limit for .xls/.xlsx files!")

	if out_ext == ".xlsx":
		try:
			with pd.ExcelWriter(path_out) as writer:
				outputs.to_excel(writer, sheet_name="output-table", engine="openpyxl")
		except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
			try:
				with pd.ExcelWriter(path_out[0:-1]) as writer:
					outputs.to_excel(writer, sheet_name="output-table", engine="xlsxwriter")
			except (ImportError, IndexError, ValueError, ModuleNotFoundError, FileNotFoundError):
				try:
					os.remove(path_out)
				except:
					pass
				outputs.to_json(os.path.splitext(path_out)[0] + ".json.zip", orient='table')
				if warn or print_log:
					click.echo(f"Fallback export to {os.path.splitext(path_out)}.json.zip due to Excel engine failure, usually caused by exceeding max row limit for .xls/.xlsx files!")
	elif out_ext == ".pkl":
		outputs.to_pickle(path_out+".xz",compression="xz")
	elif out_ext == ".xz":
		outputs.to_pickle(path_out,compression="xz")
	elif out_ext == ".json":
		outputs.to_json(path_out,orient='table')
	elif out_ext == ".csv":
		outputs.to_csv(path_out,escapechar='\\')
	elif out_ext == ".txt":
		outputs.to_string(path_out)
	elif out_ext == ".dta":
		outputs.to_stata(path_out)
	elif out_ext == ".parquet":
		outputs.to_parquet(path_out)
	else:
		pass
	return outputs 

def archive(conf,parquet_test=False): 
	"""
	Write full text archive to file.pkl.xz
	"""
	path_in = conf['INPUT_PATH']
	path_out = conf['OUTPUT_PATH']
	arc_out = conf['OUTPUT_PATH']
	out_ext = conf['OUTPUT_EXT']
	count = conf['COUNT']
	queue = conf['QUEUE']
	print_log = conf['LOG']
	warn = conf['WARN']
	no_write = conf['NO_WRITE']
	dedupe = conf['DEDUPE']
	table = conf['TABLE']
	dedupe = conf['DEDUPE']
	from_archive = True if conf['IS_FULL_TEXT']==True else False

	start_time = time.time()
	if warn == False:
		warnings.filterwarnings("ignore")
	if warn:
		click.echo(click.style("* ",blink=True) + "Creating full text archive...")

	if not from_archive:
		allpagestext = pd.Series(queue).map(lambda x: get.PDFText(x))
	else:
		allpagestext = pd.Series(queue)

	outputs = pd.DataFrame({
		'Path': queue if from_archive else np.nan,
		'AllPagesText': allpagestext,
		'Timestamp': start_time,
		})

	outputs.fillna('',inplace=True)

	if not no_write and not parquet_test:
		outputs.to_pickle(path_out,compression="xz")
	if parquet_test:
		outputs.to_parquet(path_out+".parquet",compression="brotli")
	parse.logs.complete(conf, start_time, outputs)
	return outputs



# conf 75
# in progress
# Sam Robson

import os
import sys
import glob
import re
import math
import xarray
import numpy as np
import xlrd
import openpyxl
import datetime
import pandas as pd
import time
import warnings
import click
import inspect
from alacorder import logs
from alacorder import get
from alacorder import parse
from alacorder import write


def inputs(path):
	found = 0
	is_full_text = False
	good = False
	pickle = None
	queue = pd.Series()
	
	if os.path.isdir(path): # if PDF directory -> good
		queue = pd.Series(glob.glob(path + '**/*.pdf', recursive=True))
		if queue.shape[0] > 0:
			found = len(queue)
			good = True
	elif os.path.isfile(path) and os.path.splitext(path)[1] == ".xz": # if archive -> good
		good = True
		try:
			pickle = pd.read_pickle(path,compression="xz")
			queue = pickle['AllPagesText']
			is_full_text = True
			found = len(queue)
		except:
			good = False
	else:
		good = False

	if good:
		echo = click.style(f"\nFound {found} cases in input.",fg='bright_blue',bold=True)
	else:
		echo = click.style(f"""Alacorder failed to configure input! Try again with a valid PDF directory or full text archive path, or run 'python -m alacorder --help' in command line for more details.""",fg='red',bold=True)

	out = pd.Series({
		'INPUT_PATH': path,
		'IS_FULL_TEXT': is_full_text,
		'QUEUE': queue,
		'FOUND': found,
		'GOOD': good,
		'PICKLE': pickle,
		'ECHO': echo
		})
	return out

def outputs(path):
	good = False
	make = None
	pickle = None
	exists = os.path.isfile(path)
	ext = os.path.splitext(path)[1]
	if os.path.splitext(path)[1] == ".xz": # if output is existing archive
		make = "archive"
		good = True
	elif os.path.splitext(path)[1] == ".xlsx" or os.path.splitext(path)[1] == ".xls": # if output is multiexport
		make = "multiexport"
		good = True
	elif os.path.splitext(path)[1] == ".csv" or os.path.splitext(path)[1] == ".dta" or os.path.splitext(path)[1] == ".json" or os.path.splitext(path)[1] == ".txt" or os.path.splitext(path)[1] == ".pkl":
		make = "singletable"
		good = True
	if good:
		echo = click.style(f"""Output path successfully configured for {"table" if (make == "multiexport" or make == "singletable") else "archive"} export.""",fg='bright_blue',bold=True) 
	else:
		echo = click.style(f"Alacorder failed to configure output! Try again with a valid path to a file with a supported extension, or run 'python -m alacorder --help' in command line for help.",fg='red',bold=True)

	out = pd.Series({
		'OUTPUT_PATH': path,
		'OUTPUT_EXT': ext,
		'MAKE': make,
		'GOOD': good,
		'EXISTING_FILE': exists,
		'ECHO': echo
		})
	return out

def set(inputs,outputs,count=0,table='',overwrite=False,launch=False,log=True,dedupe=False,warn=False,no_write=False,no_prompt=False,skip_echo=False,debug=False,no_batch=False):

	status_code = []
	echo = ""
	will_archive = False
	will_overwrite = False
	good = True

	if not debug:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')

	## COUNT 
	content_len = inputs['FOUND']
	if content_len > count and count != 0:
		ind = count - 1
		queue = inputs.QUEUE[0:ind] 
	else:
		queue = inputs.QUEUE


	echo += logs.echo_conf(inputs.INPUT_PATH,outputs.MAKE,outputs.OUTPUT_PATH,overwrite,no_write,dedupe,launch,warn,no_prompt)

	out = pd.Series({
		'GOOD': good,
		'ECHO': echo,
		'STATUS_CODES': status_code,

		'QUEUE': queue,
		'COUNT': count,
		'IS_FULL_TEXT': bool(inputs.IS_FULL_TEXT),
		'MAKE': outputs.MAKE,
		'TABLE': table,

		'INPUT_PATH': inputs.INPUT_PATH,
		'OUTPUT_PATH': outputs.OUTPUT_PATH,
		'OUTPUT_EXT': outputs.OUTPUT_EXT,

		'OVERWRITE': will_overwrite,
		'FOUND': inputs.FOUND,

		'DEDUPE': dedupe, # not ready (well none of its ready but especially that)
		'LOG': log,
		'WARN': warn,
		'LAUNCH': launch,
		'DEBUG': debug,
		'NO_PROMPT': no_prompt,
		'NO_WRITE': no_write,
		'NO_BATCH': no_batch
		})

	return out

def batcher(conf):
	q = conf['QUEUE']
	if conf.IS_FULL_TEXT: 
		batchsize = q.shape[0] / 2
	else: 
		batchsize = 1000
	if conf.FOUND < 1000:
		batchsize = 100
	if conf.FOUND > 10000:
		batchsize = conf.FOUND / 20
	batches = np.array_split(conf.QUEUE, math.floor(conf.FOUND/batchsize))
	return batches

# same as calling conf.set(conf.inputs(path), conf.outputs(path), **kwargs)
def setpaths(input_path, output_path, count=0, table='', overwrite=False, launch=False, log=True, dedupe=False, warn=False,no_write=False, no_prompt=False, skip_echo=False, debug=False, no_batch=False):
	if not debug:
		sys.tracebacklimit = 0
		warnings.filterwarnings('ignore')
	a = inputs(input_path)
	if log:
		click.echo(a.ECHO)
	b = outputs(output_path)
	if log:
		click.echo(b.ECHO)
	c = set(a,b, count=0, table='', overwrite=overwrite, launch=launch, log=log, dedupe=dedupe, warn=warn, no_write=no_write, no_prompt=no_prompt, debug=debug, no_batch=no_batch)
	if log:
		click.echo(c.ECHO)
	return c



